/* M2C Modula-2 Compiler & Translator
 * Copyright (c) 2015, 2016 Benjamin Kowarsch
 *
 * @synopsis
 *
 * M2C is a compiler and translator for the classic Modula-2 programming
 * language as described in the 3rd and 4th editions of Niklaus Wirth's
 * book "Programming in Modula-2" (PIM) published by Springer Verlag.
 *
 * In compiler mode, M2C compiles Modula-2 source via C to object files or
 * executables using the host system's resident C compiler and linker.
 * In translator mode, it translates Modula-2 source to C source.
 *
 * Further information at http://savannah.nongnu.org/projects/m2c/
 *
 * @file
 *
 * m2-caseconv.h
 *
 * Public interface for Modula-2 to C identifier conversion.
 *
 * @license
 *
 * M2C is free software: you can redistribute and/or modify it under the
 * terms of the GNU Lesser General Public License (LGPL) either version 2.1
 * or at your choice version 3 as published by the Free Software Foundation.
 *
 * M2C is distributed in the hope that it will be useful,  but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  Read the license for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with m2c.  If not, see <https://www.gnu.org/copyleft/lesser.html>.
 */

#ifndef M2C_IDENTCONV_H
#define M2C_IDENTCONV_H

#include "m2-common.h"

void set_cstyle_option (void);

/* --------------------------------------------------------------------------
 * function get_include_guard_ident(target, module_id)
 * --------------------------------------------------------------------------
 * Composes a C include guard macro identifier from module_id and passes it
 * as a NUL terminated C string in out-parameter target.
 *
 * includeGuard :=
 *   vModeIncludeGuard | cModeIncludeGuard
 *   ;
 *
 * vModeIncludeGuard :=
 *   'MODULE__' moduleIdent '__H'
 *   ;
 *
 * moduleIdent :=
 *   verbatim copy of module identifier
 *   ;
 *
 * cModeIncludeGuard :=
 *   upperModuleIdent '_H'
 *   ;
 *
 * upperModuleIdent :=
 *   uppercase/lowline transformation of module identifier
 *   ;
 *
 * Example:
 *   DEFINITION MODULE FooLib;
 *   [Verbatim] => MODULE__FooLib__H
 *   [C style]  => FOO_LIB_H
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if module_id is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_include_guard_ident (char_ptr_t *target, const char *module_id);


/* --------------------------------------------------------------------------
 * function get_public_const_ident(target, module_id, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified public constant identifier from module_id and ident,
 * and passes it as a NUL terminated C string in out-parameter target.
 *
 * publicConstIdent :=
 *   vModePublicConstIdent | cModePublicConstIdent
 *   ;
 *
 * vModePublicConstIdent :=
 *   publicQualifier sourceIdent
 *   ;
 *
 * publicQualifier :=
 *   moduleIdent '__'
 *   ;
 *
 * moduleIdent :=
 *   verbatim copy of module identifier
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of constant identifier |
 *   uppercase/lowline transformation of constant identifier
 *   ;
 *
 * cModePublicConstIdent :=
 *   upperPublicQualifier upperSourceIdent
 *   ;
 *
 * upperPublicQualifier :=
 *   upperModuleIdent '__'
 *   ;
 *
 * upperModuleIdent :
 *   uppercase/lowline transformation of module identifier
 *   ;
 *
 * upperSourceIdent :=
 *   uppercase/lowline transformation of constant identifier
 *   ;
 *
 * Example:
 *   DEFINITION MODULE FooLib; CONST barBaz;
 *   [Verbatim] => FooLib__barBaz
 *   [C style]  => FOO_LIB__BAR_BAZ
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_public_const_ident
  (char_ptr_t *target, const char *module_id, const char *ident);


/* --------------------------------------------------------------------------
 * function get_public_type_id(target, module_id, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified public type identifier from module_id and ident,
 * and passes it as a NUL terminated C string in out-parameter target.
 *
 * publicTypeIdent :=
 *   vModePublicTypeIdent | cModePublicTypeIdent
 *   ;
 *
 * vModePublicTypeIdent :=
 *   publicQualifier sourceIdent
 *   ;
 *
 * publicQualifier :=
 *   moduleIdent '__'
 *   ;
 *
 * moduleIdent :=
 *   verbatim copy of module identifier
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of type identifier
 *   ;
 *
 * cModePublicTypeIdent :=
 *   lowerPublicQualifier lowerSourceIdent '_t'
 *   ;
 *
 * lowerPublicQualifier :=
 *   lowerModuleIdent '__'
 *   ;
 *
 * lowerModuleIdent :
 *   lowercase/lowline transformation of module identifier
 *   ;
 *
 * lowerSourceIdent :=
 *   lowercase/lowline transformation of constant identifier
 *   ;
 *
 * Example:
 *   DEFINITION MODULE FooLib; TYPE BarBaz;
 *   [Verbatim] => FooLib__BarBaz
 *   [C style]  => foo_lib__bar_baz_t
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_public_type_ident
  (char_ptr_t *target, const char *module_id, const char *ident);


/* --------------------------------------------------------------------------
 * function get_public_vfn_ident(target, module_id, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified public variable or function identifier from module_id
 * and ident, passes it as a NUL terminated C string in out-parameter target.
 *
 * publicVarOrFnIdent :=
 *   vModePublicVarOrFnIdent | cModePublicVarOrFnIdent
 *   ;
 *
 * vModePublicVarOrFnIdent :=
 *   publicQualifier sourceIdent
 *   ;
 *
 * publicQualifier :=
 *   moduleIdent '__'
 *   ;
 *
 * moduleIdent :=
 *   verbatim copy of module identifier
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of variable/function identifier
 *   ;
 *
 * cModePublicVarOrFnIdent :=
 *   lowerPublicQualifier lowerSourceIdent
 *   ;
 *
 * lowerPublicQualifier :=
 *   lowerModuleIdent '__'
 *   ;
 *
 * lowerModuleIdent :
 *   lowercase/lowline transformation of module identifier
 *   ;
 *
 * lowerSourceIdent :=
 *   lowercase/lowline transformation of variable/function identifier
 *   ;
 *
 * Example:
 *   DEFINITION MODULE FooLib; VAR barBaz;
 *   [Verbatim] => FooLib__barBaz
 *   [C style]  => foo_lib__bar_baz
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

#define get_public_var_ident(_tgt,_id) get_public_vfn_ident(_tgt, _id)

#define get_public_fn_ident(_tgt,_id) get_public_vfn_ident(_tgt, _id)

uint_t get_public_vfn_ident
  (char_ptr_t *target, const char *module_id, const char *ident);


/* --------------------------------------------------------------------------
 * function get_private_const_ident(target, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified private constant identifier from ident and passes
 * it as a NUL terminated C string in out-parameter target.
 *
 * privateConstIdent :=
 *   vModePrivateConstIdent | cModePrivateConstIdent
 *   ;
 *
 * vModePrivateConstIdent :=
 *   privateQualifier sourceIdent
 *   ;
 *
 * privateQualifier :=
 *   'Private__'
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of constant identifier
 *
 * cModePrivateConstIdent :=
 *   upperPrivateQualifier upperSourceIdent
 *   ;
 *
 * upperPrivateQualifier :=
 *   'PRIVATE__'
 *   ;
 *
 * upperSourceIdent :=
 *   uppercase/lowline transformation of constant identifier
 *   ;
 *
 * Example:
 *   IMPLEMENTATION MODULE ...; CONST fooBar;
 *   [Verbatim] => Private__fooBar
 *   [C style]  => PRIVATE__FOO_BAR
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_private_const_ident (char_ptr_t *target, const char *ident);


/* --------------------------------------------------------------------------
 * function get_private_type_ident(target, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified private type identifier from ident and passes it as
 * a NUL terminated C string in out-parameter target.
 *
 * privateTypeIdent :=
 *   vModePrivateTypeIdent | cModePrivateTypeIdent
 *   ;
 *
 * vModePrivateTypeIdent :=
 *   privateQualifier sourceIdent
 *   ;
 *
 * privateQualifier :=
 *   'Private__'
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of type identifier
 *
 * cModePrivateTypeIdent :=
 *   lowerPrivateQualifier lowerSourceIdent '_t'
 *   ;
 *
 * lowerPrivateQualifier :=
 *   'private__'
 *   ;
 *
 * lowerSourceIdent :=
 *   lowercase/lowline transformation of type identifier
 *   ;
 *
 * Example:
 *   IMPLEMENTATION MODULE ...; TYPE FooBar;
 *   [Verbatim] => Private__FooBar
 *   [C style]  => private__foo_bar_t
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_private_type_ident (char_ptr_t *target, const char *ident);


/* --------------------------------------------------------------------------
 * function get_private_vfn_ident(target, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified private variable or function identifier from ident
 * and passes it as a NUL terminated C string in out-parameter target.
 *
 * privateVarOrFnIdent :=
 *   vModePrivateVarOrFnIdent | cModePrivateVarOrFnIdent
 *   ;
 *
 * vModePrivateVarOrFnIdent :=
 *   privateQualifier sourceIdent
 *   ;
 *
 * privateQualifier :=
 *   'Private__'
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of variable/function identifier
 *
 * cModePrivateVarOrFnIdent :=
 *   lowerPrivateQualifier lowerSourceIdent
 *   ;
 *
 * lowerPrivateQualifier :=
 *   'private__'
 *   ;
 *
 * lowerSourceIdent :=
 *   lowercase/lowline transformation of variable/function identifier
 *   ;
 *
 * Example:
 *   IMPLEMENTATION MODULE ...; VAR fooBar;
 *   [Verbatim] => Private__fooBar
 *   [C style]  => private__foo_bar
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

#define get_private_var_ident(_tgt,_id) get_private_vfn_ident(_tgt, _id)

#define get_private_fn_ident(_tgt,_id) get_private_vfn_ident(_tgt, _id)

uint_t get_private_vfn_ident (char_ptr_t *target, const char *ident);


/* --------------------------------------------------------------------------
 * function get_local_const_ident(target, outer, ordinal, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified local constant identifier from outer or ordinal and
 * ident, passes it as a NUL terminated C string in out-parameter target.
 *
 * localConstIdent :=
 *   vModeLocalConstIdent | cModeLocalConstIdent
 *   ;
 *
 * vModeLocalConstIdent :=
 *   localQualifier enclFuncQualifier sourceIdent
 *   ;
 *
 * localQualifier :=
 *   'Local__'
 *   ;
 *
 * enclFuncQualifier :=
 *   ( enclFunctionIdent | enclFunctionOrdinal ) '__'
 *   ;
 *
 * enclFunctionIdent
 *   verbatim copy of enclosing function identifier
 *   ;
 *
 * enclFunctionOrdinal :=
 *   Base10Digit Base16Digit*
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of constant identifier
 *   ;
 *
 * cModeLocalConstIdent :=
 *   upperLocalQualifier upperEnclFuncQualifier upperSourceIdent
 *   ;
 *
 * upperLocalQualifier :=
 *   'LOCAL__'
 *   ;
 *
 * upperEnclFuncQualifier :=
 *   ( upperEnclFunctionIdent | enclFunctionOrdinal ) '__'
 *   ;
 *
 * upperEnclFunctionIdent
 *   uppercase/lowline transformation of enclosing function identifier
 *   ;
 *
 * upperSourceIdent :=
 *   uppercase/lowline transformation of constant identifier
 *   ;
 *
 * Example:
 *   PROCEDURE outerProc; CONST fooBar;
 *   [Verbatim] => Local__outerProc__fooBar
 *   [C style]  => LOCAL__OUTER_PROC__FOO_BAR
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_local_const_ident
  (char_ptr_t *target, const char *outer, uint_t ordinal, const char *ident);


/* --------------------------------------------------------------------------
 * function get_local_type_ident(target, outer, ordinal, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified local type identifier from outer or ordinal and
 * ident, passes it as a NUL terminated C string in out-parameter target.
 *
 * localTypeIdent :=
 *   vModeLocalTypeIdent | cModeLocalTypeIdent
 *   ;
 *
 * vModeLocalTypeIdent :=
 *   localQualifier enclFuncQualifier sourceIdent
 *   ;
 *
 * localQualifier :=
 *   'Local__'
 *   ;
 *
 * enclFuncQualifier :=
 *   ( enclFunctionIdent | enclFunctionOrdinal ) '__'
 *   ;
 *
 * enclFunctionIdent
 *   verbatim copy of enclosing function identifier
 *   ;
 *
 * enclFunctionOrdinal :=
 *   Base10Digit Base16Digit*
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of type identifier
 *   ;
 *
 * cModeLocalConstIdent :=
 *   lowerLocalQualifier lowerEnclFuncQualifier lowerSourceIdent '_t'
 *   ;
 *
 * lowerLocalQualifier :=
 *   'local__'
 *   ;
 *
 * lowerEnclFuncQualifier :=
 *   ( lowerEnclFunctionIdent | enclFunctionOrdinal ) '__'
 *   ;
 *
 * lowerEnclFunctionIdent
 *   lowercase/lowline transformation of enclosing function identifier
 *   ;
 *
 * lowerSourceIdent :=
 *   lowercase/lowline transformation of type identifier
 *   ;
 *
 * Example:
 *   PROCEDURE outerProc; TYPE FooBar;
 *   [Verbatim] => Local__outerProc__FooBar
 *   [C style]  => local__outer_proc__foo_bar_t
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_local_type_ident
  (char_ptr_t *target, const char *outer, uint_t ordinal, const char *ident);


/* --------------------------------------------------------------------------
 * function get_local_var_ident(target, ident)
 * --------------------------------------------------------------------------
 * Composes a disambiguated local variable identifier from ident,
 * passes it as a NUL terminated C string in out-parameter target.
 *
 * localVarIdent :=
 *   vModeLocalVarIdent | cModeLocalVarIdent
 *   ;
 *
 * vModeLocalVarIdent :=
 *   sourceIdent collisionPreventionSuffix?
 *   ;
 *
 * collisionPreventionSuffix :=
 *   '_'
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of variable identifier
 *   ;
 *
 * cModeLocalVarIdent :=
 *   lowerSourceIdent collisionPreventionSuffix?
 *   ;
 *
 * lowerSourceIdent :=
 *   lowercase/lowline transformation of variable identifier
 *   ;
 *
 * Example:
 *   PROCEDURE outerProc; VAR fooBar;
 *   [Verbatim] => fooBar
 *   [C style]  => foo_bar
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_local_var_ident (char_ptr_t *target, const char *ident);


/* --------------------------------------------------------------------------
 * function get_local_fn_ident(target, outer, ordinal, ident)
 * --------------------------------------------------------------------------
 * Composes a qualified local function identifier from outer or ordinal and
 * ident, passes it as a NUL terminated C string in out-parameter target.
 *
 * localFuncIdent :=
 *   vModeLocalFuncIdent | cModeLocalFuncIdent
 *   ;
 *
 * vModeLocalFuncIdent :=
 *   localQualifier enclFuncQualifier sourceIdent
 *   ;
 *
 * localQualifier :=
 *   'Local__'
 *   ;
 *
 * enclFuncQualifier :=
 *   ( enclFunctionIdent | enclFunctionOrdinal ) '__'
 *   ;
 *
 * enclFunctionIdent
 *   verbatim copy of enclosing function identifier
 *   ;
 *
 * enclFunctionOrdinal :=
 *   Base10Digit Base16Digit*
 *   ;
 *
 * sourceIdent :=
 *   verbatim copy of enclosed function identifier
 *   ;
 *
 * cModeLocalFuncIdent :=
 *   lowerLocalQualifier lowerEnclFuncQualifier lowerSourceIdent
 *   ;
 *
 * lowerLocalQualifier :=
 *   'local__'
 *   ;
 *
 * lowerEnclFuncQualifier :=
 *   ( lowerEnclFunctionIdent | enclFunctionOrdinal ) '__'
 *   ;
 *
 * lowerEnclFunctionIdent
 *   lowercase/lowline transformation of enclosing function identifier
 *   ;
 *
 * lowerSourceIdent :=
 *   lowercase/lowline transformation of enclosed function identifier
 *   ;
 *
 * Example:
 *   PROCEDURE outerProc; PROCEDURE innerProc;
 *   [Verbatim] => Local__outerProc__innerProc
 *   [C style]  => local__outer_proc__inner_proc
 *
 * Returns length of composed identifier on success.
 * Fails and returns zero if ident is NULL or empty.
 * ----------------------------------------------------------------------- */

uint_t get_local_fn_ident
  (char_ptr_t *target, const char *outer, uint_t ordinal, const char *ident);


#endif /* M2C_IDENTCONV_H */

/* END OF FILE */